/* wmudmount
 * Copyright © 2010  Brad Jorsch <anomie@users.sourceforge.net>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include "config.h"

#include <string.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/stat.h>
#include <sys/types.h>
#ifdef HAVE_FSTAB_FUNCS
#include <fstab.h>
#endif
#include <dbus/dbus.h>
#include <dbus/dbus-glib.h>
#include <dbus/dbus-glib-lowlevel.h>

#include "asklabel.h"
#include "askpass.h"
#include "die.h"
#include "dock.h"
#include "fsm.h"
#include "misc.h"
#include "udisks.h"

const char *udisks_daemon = NULL;
int num_volumes = 0;
static int max_volumes = 4;
udisks_volume **Volumes = NULL;
static int num_all_volumes = 0;
static int max_all_volumes = 4;
static udisks_volume **all_volumes = NULL;

static DBusGConnection *conn = NULL;
static DBusGProxy *connproxy = NULL, *proxy = NULL, *propproxy = NULL;

#define DKD_009 1
#define UDISKS_1_0 2

static gboolean in_connect = FALSE;
static int daemon_version = 0;
static const char *dbus_name = NULL;
static const char *dbus_dev_interface = NULL;

/* Data */

static int num_fs_info=0;
struct udisks_fs_info **fs_info = NULL;
static const struct udisks_fs_info *get_fs_info(const char *id);

/* The following can be used to sort the lists correctly:
 * perl -nwe 'BEGIN { %x=(); } $x{"$1\xff"}=$_ if /^[^"]*"([^"]*)"/; END { for $k (sort keys %x){ print $x{$k}; } }'
 */
struct disk_info {
    const char *id;
    const char *desc;
    const char *icon;
    const char *icon_r;
    gboolean always_readonly;
};
static struct disk_info media_info[] = {
    {"flash_cf",              "CompactFlash",   "media-flash-cf",              NULL, 0},
    {"flash_mmc",             "MMC",            "media-flash-sd-mmc",          NULL, 0},
    {"flash_ms",              "MemoryStick",    "media-flash-ms",              NULL, 0},
    {"flash_sdhc",            "SDHC",           "media-flash-sd-hc",           NULL, 0},
    {"flash_sd",              "SD",             "media-flash-sd",              NULL, 0},
    {"flash_sm",              "SmartMedia",     "media-flash-sm",              NULL, 0},
    {"flash",                 "Flash Card",     "media-flash-sd",              NULL, 0},
    {"floppy_jaz",            "Jaz Disk",       "media-floppy-jaz",            NULL, 0},
    {"floppy_zip",            "Zip Disk",       "media-floppy-zip",            NULL, 0},
    {"floppy",                "Floppy",         "media-floppy",                NULL, 0},
    {"optical_bd_re",         "Blu-Ray RW",     "media-optical-bd-re",         NULL, 0},
    {"optical_bd_r",          "Blu-Ray R",      "media-optical-bd-r",          NULL, 0},
    {"optical_bd",            "Blu-Ray",        "media-optical-bd-rom",        NULL, 1},
    {"optical_cd_rw",         "CD-RW",          "media-optical-cd-rw",         NULL, 0},
    {"optical_cd_r",          "CD-R",           "media-optical-cd-r",          NULL, 0},
    {"optical_cd",            "CD-ROM",         "media-optical-cd-rom",        NULL, 1},
    {"optical_dvd_plus_r_dl", "DVD+R DL",       "media-optical-dvd-dl-r-plus", NULL, 0},
    {"optical_dvd_plus_rw_dl","DVD+RW DL",      "media-optical-dvd-dl-rw-plus",NULL, 0},
    {"optical_dvd_plus_rw",   "DVD+RW",         "media-optical-dvd-rw-plus",   NULL, 0},
    {"optical_dvd_plus_r",    "DVD+R",          "media-optical-dvd-r-plus",    NULL, 0},
    {"optical_dvd_ram",       "DVD-RAM",        "media-optical-dvd-ram",       NULL, 0},
    {"optical_dvd_rw",        "DVD-RW",         "media-optical-dvd-rw",        NULL, 0},
    {"optical_dvd_r",         "DVD-R",          "media-optical-dvd-r",         NULL, 0},
    {"optical_dvd",           "DVD-ROM",        "media-optical-dvd-rom",       NULL, 1},
    {"optical_hddvd_rw",      "HD DVD-RW",      "media-optical-hddvd-rw",      NULL, 0},
    {"optical_hddvd_r",       "HD DVD-R",       "media-optical-hddvd-r",       NULL, 0},
    {"optical_hddvd",         "HD DVD",         "media-optical-hddvd-rom",     NULL, 1},
    {"optical_mo",            "Magneto Optical","media-optical-mo",            NULL, 0},
    {"optical_mrw_w",         "Mount Ranier/W", "media-optical-mrw-w",         NULL, 0},
    {"optical_mrw",           "Mount Ranier",   "media-optical-mrw",           NULL, 1},
    {"optical",               "Optical Disk",   "media-optical",               NULL, 0},
};
static struct disk_info iface_info[] = {
    {"ata_parallel",          "PATA Disk",      "drive-harddisk-ata-parallel",    "drive-removable-media-ata-parallel",     0},
    {"ata_serial_esata",      "eSATA Disk",     "drive-harddisk-ata-serial-esata","drive-removable-media-ata-serial-esata", 0},
    {"ata_serial",            "SATA Disk",      "drive-harddisk-ata-serial",      "drive-removable-media-ata-serial",       0},
    {"ata",                   "ATA Disk",       "drive-harddisk-ata",             "drive-removable-media-ata",              0},
    {"firewire",              "Firewire Disk",  "drive-harddisk-ieee1394",        "drive-removable-media-ieee1394",         0},
    {"platform",              "Platform Disk",  "drive-harddisk-platform",        "drive-removable-media-platform",         0},
    {"scsi",                  "SCSI Disk",      "drive-harddisk-scsi",            "drive-removable-media-scsi",             0},
    {"sdio",                  "SDIO Disk",      "drive-harddisk-sdio",            "drive-removable-media-sdio",             0},
    {"usb",                   "USB Disk",       "drive-harddisk-usb",             "drive-removable-media-usb",              0},
    {"virtual",               "Virtual Disk",   "drive-harddisk-virtual",         "drive-removable-media-virtual",          0},
};

/* Volume funcs */

static udisks_volume *alloc_volume(const char *path);
static gboolean update_volume(udisks_volume *v);
static void device_changed(DBusGProxy *proxy, const char *path, gpointer data);

static gboolean insert_volume(udisks_volume *v){
    for(int i=0; i<num_volumes; i++){
        if(Volumes[i]==v) return FALSE;
    }
    warn(DEBUG_DEBUG, "Adding volume %s (%p) to active list", v->path, v);
    if(num_volumes >= max_volumes){
        max_volumes *= 2;
        Volumes = g_realloc(Volumes, max_volumes * sizeof(*Volumes));
    }
    Volumes[num_volumes++] = v;
    volume_added(v, in_connect);
    return TRUE;
}

static gboolean remove_volume(udisks_volume *v){
    for(int i=num_volumes-1; i>=0; i--){
        if(Volumes[i]==v){
            num_volumes--;
            g_memmove(Volumes+i, Volumes+i+1, (num_volumes-i)*sizeof(udisks_volume *));
            warn(DEBUG_DEBUG, "Removing volume %s (%p) at index %d from active list", v->path, v, i);
            volume_removed(i, v, in_connect);
            return TRUE;
        }
    }
    return FALSE;
}

static gboolean is_fstab_changed(void){
    static time_t fstab_mtime = 0;
    struct stat s;
    gboolean ret = FALSE;
    if(!stat("/etc/fstab", &s)){
        ret = (s.st_mtime != fstab_mtime);
        fstab_mtime = s.st_mtime;
    }
    return ret;
}

gboolean fsm_blink = FALSE;
static gboolean poll_blocks(gpointer p){
    gboolean update = FALSE;
    udisks_volume *v;
    for(int i=0; i<num_volumes; i++){
        v = Volumes[i];
        if(v->type != UDISKS_FILESYSTEM) continue;
        if(v->d.fs.mountpoint){
            guint64 old_b = v->d.fs.total_size;
            guint64 old_f = v->d.fs.free_size;
            guint64 b, f;
            double pct;
            if(get_fs_usage(v->d.fs.mountpoint, &b, &f) && b){
                v->d.fs.total_size = b;
                v->d.fs.free_size = f;
                pct = (b - f) / (double)b;
            } else {
                v->d.fs.total_size = 0;
                v->d.fs.free_size = 0;
                pct = -1;
            }
            if(b != old_b || f != old_f) update = TRUE;
            if(pct >= BLINK_THRESHOLD) update = TRUE;
        } else {
            v->d.fs.total_size = 0;
            v->d.fs.free_size = 0;
        }
    }
    fsm_blink = !fsm_blink;
    if(update) fsm_updated();

    if(is_fstab_changed()){
        warn(DEBUG_DEBUG, "/etc/fstab changed!");
        for(int i = num_volumes-1; i>=0; i--){
            v = Volumes[i];
            if(v->type == UDISKS_FILESYSTEM) device_changed(NULL, v->path, NULL);
        }
        redraw_dock();
    }

    return TRUE;
}

static udisks_volume *alloc_volume(const char *path){
    warn(DEBUG_DEBUG, "Allocating new volume for %s", path);
    udisks_volume *v = g_slice_new(udisks_volume);
    v->refct = 1;
    v->path = g_strdup(path);
    v->parent = NULL;
    v->proxy = dbus_g_proxy_new_for_name(conn, dbus_name, path, "org.freedesktop.DBus.Properties");
    if(!v->proxy) die("Could not create device proxy for %s", path);
    v->type = 0;
    v->native_path = NULL;
    v->label = NULL;
    v->old_label = NULL;
    v->desc = NULL;
    v->dev = NULL;
    v->uuid = NULL;
    v->icon = NULL;
    v->unemblemed_icon = NULL;
    v->pending_actions = 0;

    if(num_all_volumes >= max_all_volumes){
        max_all_volumes *= 2;
        all_volumes = g_realloc(all_volumes, max_all_volumes * sizeof(*all_volumes));
    }
    all_volumes[num_all_volumes++] = v;
    if(!update_volume(v)){
        udisks_unref(v);
        return NULL;
    }
    return v;
}

static void free_volume_data(udisks_volume *v){
    v->parent = NULL;
    g_free((void *)v->dev); v->dev = NULL;
    g_free((void *)v->native_path); v->native_path = NULL;
    g_free((void *)v->uuid); v->uuid = NULL;
    g_free((void *)v->label); v->label = NULL;
    g_free((void *)v->desc); v->desc = NULL;
    if(v->icon){ g_object_unref(v->icon); v->icon = NULL; }
    if(v->unemblemed_icon){ g_object_unref(v->unemblemed_icon); v->unemblemed_icon = NULL; }
    switch(v->type){
      case UDISKS_FILESYSTEM:
        g_free((void *)v->d.fs.mountpoint); v->d.fs.mountpoint = NULL;
        break;
      case UDISKS_LUKS:
        g_free((void *)v->d.luks.device_uuid); v->d.luks.device_uuid = NULL;
        break;
    }
    v->type = 0;
}

void udisks_ref(udisks_volume *v){
    v->refct++;
}

static void rm_volume(udisks_volume *v){
    int i;
    remove_volume(v);
    for(i=num_all_volumes-1; i>=0; i--){
        if(all_volumes[i]->parent == v){
            if(!in_connect) warn(DEBUG_WARN, "Device %s (%p) being freed before child device %s", v->path, v, all_volumes[i]->path);
            udisks_volume *p = all_volumes[i];
            rm_volume(p);
            udisks_unref(p);
            i=num_all_volumes; // restart scan, since array may have changed
        }
    }
    for(i=num_all_volumes-1; i>=0; i--){
        if(all_volumes[i] == v){
            num_all_volumes--;
            g_memmove(all_volumes+i, all_volumes+i+1, (num_all_volumes-i)*sizeof(udisks_volume *));
        }
    }
}
void udisks_unref(udisks_volume *v){
    if(--v->refct > 0) return;

    rm_volume(v);
    free_volume_data(v);
    g_free((void *)v->path);
    g_free((void *)v->old_label);
    if(v->proxy) g_object_unref(v->proxy);
    g_slice_free(udisks_volume, v);
}

static gboolean checkbool(GHashTable *t, const char *key, gboolean dflt){
    GValue *value = g_hash_table_lookup(t, key);
    return value?g_value_get_boolean(value):dflt;
}

static const char *getstr(GHashTable *t, const char *key){
    GValue *value = g_hash_table_lookup(t, key);
    if(!value) return NULL;
    const char *c = g_value_get_string(value);
    return (c && *c)?c:NULL;
}

static const char *getobjectpath(GHashTable *t, const char *key){
    GValue *value = g_hash_table_lookup(t, key);
    if(!value) return NULL;
    const char *c = (const char *)g_value_get_boxed(value);
    return (c && *c && !(c[0]=='/' && c[1]=='\0'))?c:NULL;
}

static gboolean update_volume(udisks_volume *v){
    static char buf[1024];
    GError *err = NULL;
    GHashTable *props;
    GValue *value;
    const char *c;

    warn(DEBUG_DEBUG, " ... Updating volume %s (%p)", v->path, v);

    dbus_g_proxy_set_interface(v->proxy, "org.freedesktop.DBus.Properties");
    if(!dbus_g_proxy_call(v->proxy, "GetAll", &err,
            G_TYPE_STRING, dbus_dev_interface,
            G_TYPE_INVALID,
            dbus_g_type_get_map("GHashTable", G_TYPE_STRING, G_TYPE_VALUE), &props,
            G_TYPE_INVALID)){
        warn(DEBUG_ERROR, "Failed to get properties for %s (%p): %s", v->path, v, err->message);
        g_error_free(err);
        return FALSE;
    }

    // Free the old values
    warn(DEBUG_DEBUG, "      ... Freeing old values");
    free_volume_data(v);
    v->root_only=0;
    v->can_unlock=0;
    v->can_lock=0;
    v->can_eject=0;
    v->can_detach=0;
    v->can_mount=0;
    v->can_relabel=0;

    // Find parent device, if any
    if(checkbool(props, "DeviceIsPartition", FALSE)){
        const char *c=getobjectpath(props, "PartitionSlave");
        if(c){
            warn(DEBUG_DEBUG, "      ... Partition is a child of %s", c);
            for(int i = 0; i<num_all_volumes; i++){
                if(!strcmp(all_volumes[i]->path, c)){
                    v->parent = all_volumes[i];
                    break;
                }
            }
            if(!v->parent){ // Huh?
                warn(DEBUG_DEBUG, "No record of %s found, adding it new", c);
                v->parent = alloc_volume(c);
                if(!v->parent){ // WTF?
                    warn(DEBUG_ERROR, "Partition %s claims to be a child of %s, but that device cannot be found", v->path, c);
                } else {
                    if(v->parent->type) insert_volume(v->parent);
                }
            }
        }
    }
    if(!v->parent && checkbool(props, "DeviceIsLuksCleartext", FALSE)){
        const char *c=getobjectpath(props, "LuksCleartextSlave");
        if(c){
            warn(DEBUG_DEBUG, "      ... Luks cleartext device is a child of %s", c);
            for(int i = 0; i<num_all_volumes; i++){
                if(!strcmp(all_volumes[i]->path, c)){
                    v->parent = all_volumes[i];
                    break;
                }
            }
            if(!v->parent){ // Huh?
                warn(DEBUG_DEBUG, "No record of %s found, adding it new", c);
                v->parent = alloc_volume(c);
                if(!v->parent){ // WTF?
                    warn(DEBUG_ERROR, "Luks cleartext device %s claims to be a child of %s, but that device cannot be found", v->path, c);
                } else {
                    if(v->parent->type) insert_volume(v->parent);
                }
            }
        }
    }

    // Load basic values
    warn(DEBUG_DEBUG, "      ... Loading new values");
    v->dev = g_strdup(getstr(props, "DeviceFile"));
    warn(DEBUG_DEBUG, "      ... Device file is %s", v->dev);
    v->native_path = g_strdup(getstr(props, "NativePath"));
    warn(DEBUG_DEBUG, "      ... Native path is %s", v->native_path);
    v->label = g_strdup(getstr(props, "DevicePresentationName"));
    if(v->label) warn(DEBUG_DEBUG, "      ... Presentation name is %s", v->label);
    const char *icon_name = getstr(props, "DevicePresentationIconName");
    if(icon_name) warn(DEBUG_DEBUG, "      ... Presentation icon is %s", icon_name);
    const char *icon_emblem = NULL;

    v->uuid = g_strdup(getstr(props, "IdUuid"));
    if(!v->uuid && checkbool(props, "DeviceIsPartition", FALSE))
        v->uuid = g_strdup(getstr(props, "PartitionUuid"));
    warn(DEBUG_DEBUG, "      ... Device UUID is %s", v->uuid);

    if(!v->label) v->label = g_strdup(getstr(props, "IdLabel"));
    if(!v->label && checkbool(props, "DeviceIsPartition", FALSE))
        v->label = g_strdup(getstr(props, "PartitionLabel"));
    if(v->label) warn(DEBUG_DEBUG, "      ... Label is %s", v->label);

    gboolean removable = checkbool(props, "DeviceIsRemovable", FALSE);
    gboolean readonly = checkbool(props, "DeviceIsReadOnly", FALSE);
    warn(DEBUG_DEBUG, "      ... Device is %sremovable", removable?"":"not ");
    warn(DEBUG_DEBUG, "      ... Device is %s", readonly?"read-only":"writable");

    // Determine icon and device description
    const char *icon = NULL;
    const char *devdesc = NULL;
    if(checkbool(props, "DeviceIsDrive", FALSE)){
        v->can_eject = checkbool(props, "DriveIsMediaEjectable", FALSE) && checkbool(props, "DeviceIsMediaAvailable", TRUE);
        v->can_detach = checkbool(props, "DriveCanDetach", FALSE);
        warn(DEBUG_DEBUG, "      ... Drive can%s eject media", v->can_eject?"":"not");
        warn(DEBUG_DEBUG, "      ... Drive can%s be detached", v->can_detach?"":"not");
        const char *media = getstr(props, "DriveMedia");
        const char *iface = getstr(props, "DriveConnectionInterface");
        gboolean done = FALSE;
        warn(DEBUG_DEBUG, "      ... Media: %s  Interface: %s", media, iface);
        if(media){
            for(int i=0; !done && i<sizeof(media_info)/sizeof(media_info[0]); i++){
                if(!g_str_has_prefix(media, media_info[i].id)) continue;
                if(!devdesc) devdesc = media_info[i].desc;
                if(!icon) icon = (removable && media_info[i].icon_r) ? media_info[i].icon_r : media_info[i].icon;
                if(media_info[i].always_readonly) readonly = FALSE;
                done = TRUE;
            }
        }
        if(iface){
            for(int i=0; !done && i<sizeof(iface_info)/sizeof(iface_info[0]); i++){
                if(!g_str_has_prefix(iface, iface_info[i].id)) continue;
                if(!devdesc) devdesc = iface_info[i].desc;
                if(!icon) icon = (removable && iface_info[i].icon_r) ? iface_info[i].icon_r : iface_info[i].icon;
                if(media_info[i].always_readonly) readonly = FALSE;
                done = TRUE;
            }
        }
    }
    if(checkbool(props, "DeviceIsLuksCleartext", FALSE)){
        icon_emblem = "wmudmount-unlock";
    }
    if(!icon_name) icon_name = icon;

    const char *usage = getstr(props, "IdUsage");
    const char *type = getstr(props, "IdType");
    if(!usage) usage = "";
    if(!type) type = "";
    if(checkbool(props, "DevicePresentationHide", FALSE)){
        warn(DEBUG_DEBUG, "      ... Device is hidden");
    } else if(!strcmp(usage, "filesystem")){
        warn(DEBUG_DEBUG, "      ... Device is a %s filesystem", type);
        const struct udisks_fs_info *fs = get_fs_info(type);
        v->type = UDISKS_FILESYSTEM;
        v->d.fs.mountpoint = NULL;
        v->d.fs.total_size = -1;
        v->d.fs.free_size = -1;
        v->d.fs.info = fs;
        v->can_mount = fs?fs->can_mount:FALSE;

        // The default rules allow anyone to mount a non-internal filesystem,
        // but only root to mount an internal one. Unless the filesystem is
        // listed in /etc/fstab, in which case it all depends on the mount
        // options in fstab. So first we check internal or not, then overwrite
        // with the fstab setting if any is found.
        v->root_only = checkbool(props, "DeviceIsSystemInternal", FALSE);
        warn(DEBUG_DEBUG, "      ... Device is %s", v->root_only?"internal":"external");
#ifdef HAVE_FSTAB_FUNCS
        // Check fstab for the "user" or "users" option. Go through all
        // fstab entries, since I've heard the last match wins.
        setfsent();
        struct fstab *e;
        while(e = getfsent()){
            char *dev, *s;
            if(g_str_has_prefix(e->fs_spec,"UUID=")){
                s = g_strdup_printf("/dev/disk/by-uuid/%s", e->fs_spec+5);
                dev = realpath(s, NULL);
                g_free(s);
            } else if(g_str_has_prefix(e->fs_spec,"LABEL=")){
                s = g_strdup_printf("/dev/disk/by-label/%s", e->fs_spec+6);
                dev = realpath(s, NULL);
                g_free(s);
            } else {
                dev = realpath(e->fs_spec, NULL);
            }
            if(!dev) continue;
            if(!strcmp(v->dev, dev)){
                warn(DEBUG_DEBUG, "      ... fstab entry for %s matches", e->fs_spec);
                v->root_only = TRUE;
                char *p, *s = NULL;
                int q=0;
                for(p=e->fs_mntops; p && *p; p++){
                    if(!s) s=p;
                    if(*p == '"') q ^= 1;
                    if(q) continue; /* In quoted string */
                    if(*p == ',') *p = '\0';
                    if(*p == '\0' || *(p+1) == '\0'){
                        if(!strcmp(s, "user") || !strcmp(s, "users")){
                            warn(DEBUG_DEBUG, "      ... Found '%s' option in fstab", s);
                            v->root_only = FALSE;
                        }
                        if(!strcmp(s, "nouser") || !strcmp(s, "defaults")){
                            warn(DEBUG_DEBUG, "      ... Found '%s' option in fstab", s);
                            v->root_only = TRUE;
                        }
                        s = NULL;
                    }
                }
            }
            free(dev);
        }
        endfsent();
#else
        warn(DEBUG_DEBUG, "      ... (cannot check fstab, support not compiled in)");
#endif

        if(checkbool(props, "DeviceIsMounted", FALSE)){
            v->can_relabel = fs?fs->supports_online_label_rename:FALSE;
            value = g_hash_table_lookup(props, "DeviceMountPaths");
            if(value){
                char **p = g_value_get_boxed(value);
                if(p && p[0] && p[0][0]){
                    icon_emblem = "emblem-mounted";
                    v->d.fs.mountpoint = g_strdup(p[0]);
                    guint64 b, f;
                    if(get_fs_usage(v->d.fs.mountpoint, &b, &f) && b){
                        v->d.fs.total_size = b;
                        v->d.fs.free_size = f;
                    }
                }
            }
            warn(DEBUG_DEBUG, "      ... Device is mounted at %s", v->d.fs.mountpoint);
        } else {
            v->can_relabel = fs?fs->supports_label_rename:FALSE;
        }
        warn(DEBUG_DEBUG, "      ... Can mount: %s  Can relabel: %s  Root only: %s", v->can_mount?"Yes":"No", v->can_relabel?"Yes":"No", v->root_only?"Yes":"No");

        buf[0] = '\0';
        int l = 0;
        value = g_hash_table_lookup(props, "DeviceSize");
        if(value){
            guint64 size = g_value_get_uint64(value);
            if(size > 0){
                c = format_byte_size(size);
                l += g_strlcpy(buf+l, c, sizeof(buf)-l);
                g_free((void *)c);
            }
        }
        c = type;
        if(fs && fs->name && *fs->name) c=fs->name;
        if(c){
            if(l) buf[l++]=' ';
            l += g_strlcpy(buf+l, c, sizeof(buf)-l);
        }
        if(l) buf[l++] = ' ';
        l += g_strlcpy(buf+l, "Filesystem", sizeof(buf)-l);
        if(devdesc){
            if(l) l += g_strlcpy(buf+l, " on ", sizeof(buf)-l);
            l += g_strlcpy(buf+l, devdesc, sizeof(buf)-l);
        }
        if(l) v->desc = g_strdup(buf);
    } else if(!strcmp(usage, "crypto")){
        icon_emblem = "wmudmount-lock";
        if(!strcmp(type, "crypto_LUKS")){
            v->d.luks.device_type = "LUKS";
            v->d.luks.device_uuid = g_strdup(getstr(props, "IdUuid"));
            warn(DEBUG_DEBUG, "      ... Is a Luks crypto device (uuid %s)", v->d.luks.device_uuid);
            c = getobjectpath(props, "LuksHolder");
            if(c){
                v->can_lock = TRUE;
                warn(DEBUG_DEBUG, "      ... Is already open (%s), don't include it in the volume list", c);
            } else {
                v->can_unlock = TRUE;
                warn(DEBUG_DEBUG, "      ... Is not yet open, include it in the volume list");
                v->type = UDISKS_LUKS;
            }
            buf[0] = '\0';
            int l = 0;
            l += g_strlcpy(buf+l, "LUKS crypto device", sizeof(buf)-l);
            if(devdesc){
                if(l) l += g_strlcpy(buf+l, " on ", sizeof(buf)-l);
                l += g_strlcpy(buf+l, devdesc, sizeof(buf)-l);
            }
            if(l) v->desc = g_strdup(buf);
        } else {
            warn(DEBUG_DEBUG, "      ... Unknown type of crypto device (type %s)", type);
        }
    } else {
        warn(DEBUG_DEBUG, "      ... Unknown device usage '%s'", usage);
    }

    if(!v->desc) v->desc = g_strdup(devdesc);
    if(v->desc) warn(DEBUG_DEBUG, "      ... Description is %s", v->desc);

    if(icon_name){
        v->icon = g_themed_icon_new_with_default_fallbacks(icon_name);
    } else {
        for(udisks_volume *p=v->parent; p; p=p->parent){
            if(p->icon){
                g_object_ref(p->unemblemed_icon);
                v->icon=p->unemblemed_icon;
                break;
            }
        }
        if(!v->icon){
            v->icon = g_themed_icon_new_with_default_fallbacks(removable ? "drive-removable-media" : "drive-harddisk");
        }
    }
    g_object_ref(v->icon);
    v->unemblemed_icon = v->icon;

    if(readonly){
        GIcon *oldicon = v->icon;
        GIcon *ro_icon = g_themed_icon_new("emblem-readonly");
        GEmblem *emblem = g_emblem_new_with_origin(ro_icon, G_EMBLEM_ORIGIN_LIVEMETADATA);
        v->icon = g_emblemed_icon_new(oldicon, emblem);
        g_object_unref(oldicon);
        g_object_unref(ro_icon);
        g_object_unref(emblem);
    }

    if(icon_emblem){
        GIcon *oldicon = v->icon;
        GIcon *emblem_icon = g_themed_icon_new(icon_emblem);
        GEmblem *emblem = g_emblem_new_with_origin(emblem_icon, G_EMBLEM_ORIGIN_DEVICE);
        v->icon = g_emblemed_icon_new(oldicon, emblem);
        g_object_unref(oldicon);
        g_object_unref(emblem_icon);
        g_object_unref(emblem);
    }

    if(v->icon){
        c = g_icon_to_string(v->icon);
        warn(DEBUG_DEBUG, "      ... Icon is %s", c);
        g_free((void *)c);
    }

    g_hash_table_unref(props);
    return TRUE;
}

/* Event handlers */

static void device_removed(DBusGProxy *proxy, const char *path, gpointer data){
    warn(DEBUG_INFO, "Device removed: %s", path);
    for(int i = num_all_volumes-1; i>=0; i--){
        if(!strcmp(all_volumes[i]->path, path)){
            udisks_volume *p = all_volumes[i];
            rm_volume(p);
            udisks_unref(p);
            return;
        }
    }
}

static void device_changed(DBusGProxy *proxy, const char *path, gpointer data){
    warn(DEBUG_INFO, "Device added/changed: %s", path);
    udisks_volume *v;
    for(int i = num_all_volumes-1; i>=0; i--){
        v = all_volumes[i];
        if(!strcmp(v->path, path)){
            warn(DEBUG_DEBUG, "Found existing node %p, updating it", v);
            if(update_volume(v)){
                if(!(v->type?insert_volume(v):remove_volume(v)))
                    volume_updated(v);
            } else {
                rm_volume(v);
                udisks_unref(v);
            }
            return;
        }
    }
    warn(DEBUG_DEBUG, "No record of %s found, adding it new", path);
    v = alloc_volume(path);
    if(v && v->type) insert_volume(v);
}

/* Interface funcs */

static const struct udisks_fs_info *get_fs_info(const char *id){
    int i;
    for(i=0; i<num_fs_info; i++){
        if(!g_strcmp0(fs_info[i]->id, id)){
            if(!fs_info[i]->is_known)
                warn(DEBUG_WARN, "Filesystem type %s is unknown", id);
            return fs_info[i];
        }
    }
    warn(DEBUG_WARN, "Filesystem type %s is unknown", id);
    i = num_fs_info++;
    fs_info = g_realloc(fs_info, sizeof(struct udisks_fs_info *)*num_fs_info);
    fs_info[i] = g_malloc(sizeof(struct udisks_fs_info));
    fs_info[i]->is_known = FALSE;
    fs_info[i]->id = g_strdup(id);
    fs_info[i]->name = g_strdup(id);
    fs_info[i]->can_mount = TRUE; // Guess
    fs_info[i]->max_label_len = 0;
    fs_info[i]->supports_label_rename = FALSE;
    fs_info[i]->supports_online_label_rename = FALSE;
    return fs_info[i];
}

static gboolean load_fs_info(void){
    warn(DEBUG_DEBUG, "Loading known filesystems from %s", dbus_name);
    GError *err = NULL;
    GValue value = {0};

    // Get array of known filesystems
    if(!dbus_g_proxy_call(propproxy, "Get", &err,
            G_TYPE_STRING, dbus_name,
            G_TYPE_STRING, "KnownFilesystems",
            G_TYPE_INVALID,
            G_TYPE_VALUE, &value,
            G_TYPE_INVALID)){
        warn(DEBUG_WARN, "Failed to fetch known filesystems: %s", err?err->message:"<unknown error>");
        if(err){ g_error_free(err); err = NULL; }
        return FALSE;
    }

    // Reallocate our array of structs, freeing any malloced values from the
    // old array
    GPtrArray *known_fs = g_value_get_boxed(&value);
    for(int i=0; i<num_fs_info; i++){
        g_free(fs_info[i]->id);
        g_free(fs_info[i]->name);
        g_free(fs_info[i]);
    }
    fs_info = g_realloc(fs_info, sizeof(struct udisks_fs_info *)*known_fs->len);
    num_fs_info = known_fs->len;
 
    // Create the type for parsing the filesystem info
    GType type;
    switch(daemon_version){
      case DKD_009:
      case UDISKS_1_0:
        type = dbus_g_type_get_struct("GValueArray", G_TYPE_STRING, G_TYPE_STRING, G_TYPE_BOOLEAN, G_TYPE_BOOLEAN, G_TYPE_BOOLEAN, G_TYPE_UINT, G_TYPE_BOOLEAN, G_TYPE_BOOLEAN, G_TYPE_BOOLEAN, G_TYPE_BOOLEAN, G_TYPE_BOOLEAN, G_TYPE_BOOLEAN, G_TYPE_BOOLEAN, G_TYPE_BOOLEAN, G_TYPE_INVALID);
        break;
      default:
        die("CODE ERROR: load_fs_info needs updating to handle daemon version %d", daemon_version);
    }

    // Actually load the data
    for(int i=0; i<known_fs->len; i++){
        fs_info[i] = g_malloc(sizeof(struct udisks_fs_info));
        fs_info[i]->is_known = TRUE;
        GValue e = {0};
        g_value_init(&e,type);
        g_value_set_static_boxed(&e,known_fs->pdata[i]);
        switch(daemon_version){
          case DKD_009:
          case UDISKS_1_0:
            dbus_g_type_struct_get(&e,
                                   0, &fs_info[i]->id,
                                   1, &fs_info[i]->name,
                                   //2, &fs_info[i]->supports_unix_owners,
                                   3, &fs_info[i]->can_mount,
                                   //4, &fs_info[i]->can_create,
                                   5, &fs_info[i]->max_label_len,
                                   6, &fs_info[i]->supports_label_rename,
                                   7, &fs_info[i]->supports_online_label_rename,
                                   //8, &fs_info[i]->supports_fsck,
                                   //9, &fs_info[i]->supports_online_fsck,
                                   //10, &fs_info[i]->supports_resize_enlarge,
                                   //11, &fs_info[i]->supports_online_resize_enlarge,
                                   //12, &fs_info[i]->supports_resize_shrink,
                                   //13, &fs_info[i]->supports_online_resize_shrink,
                                   G_MAXUINT);
            break;
          default:
            die("CODE ERROR: load_fs_info needs updating to handle daemon version %d", daemon_version);
        }
    }

    g_value_unset(&value);
    return TRUE;
}

// This all is a little complicated. Basically, we call try_to_connect on init
// to make the initial connection. If either the dbus connection dies or the
// udisks connection dies or changes owner, we get into destroy_signaled() and
// reconnect. If we're not connected and we get a NameOwnerChanged for udisks,
// we try to connect. And most dbus failures set a timer to retry.
static gboolean try_to_connect(gpointer p);
static void destroy_signaled(GObject *sender, gpointer data);
int connect_err_level = DEBUG_ERROR;
static void destroy_signaled(GObject *sender, gpointer data){
    gboolean is_conn = (sender == G_OBJECT(connproxy));
    warn(DEBUG_ERROR, is_conn?"DBus connection lost":"%s connection lost", udisks_daemon);

    in_connect = TRUE;
    while(num_all_volumes>0) rm_volume(all_volumes[0]);
    if(proxy){
        g_signal_handlers_disconnect_by_func(proxy, G_CALLBACK(destroy_signaled), NULL);
        g_object_unref(proxy); proxy=NULL;
    }
    if(propproxy) g_object_unref(propproxy); propproxy=NULL;
    if(conn && is_conn){
        if(connproxy){
            g_signal_handlers_disconnect_by_func(connproxy, G_CALLBACK(destroy_signaled), NULL);
            g_object_unref(connproxy); connproxy=NULL;
        }
        dbus_g_connection_unref(conn); conn = NULL;
    }
    if(udisks_daemon){
        g_free((void *)udisks_daemon);
        udisks_daemon = NULL;
        udisks_connection_changed(NULL);
    }
    daemon_version = 0;
    in_connect = FALSE;
    g_timeout_add_seconds(1, try_to_connect, NULL);
}
static void name_owner_changed(DBusGProxy *proxy, const char *name, const char *old_owner, const char *new_owner, gpointer data){
    // If a new owner showed up and we're not already connected, try to connect
    // to it.
    warn(DEBUG_DEBUG, "NameOwnerChanged for %s: %s => %s", name, old_owner, new_owner);
    if((!proxy && !g_strcmp0(name, "org.freedesktop.UDisks")) || (daemon_version<UDISKS_1_0 && !g_strcmp0(name, "org.freedesktop.DeviceKit.Disks")))
        try_to_connect(NULL);
}
static gboolean try_to_connect(gpointer p){
    if(proxy) return FALSE; // already connected

    in_connect = TRUE;

    GError *err = NULL;
    GValue value = {0};
    const char *ver, *path;
    char *daemon = NULL;
    gboolean ok;

    if(!conn){
        warn(DEBUG_DEBUG, "Connecting to dbus");
        conn = dbus_g_bus_get(DBUS_BUS_SYSTEM, &err);
        if(err){
            conn = NULL;
            warn(connect_err_level, "Failed to connect to dbus: %s", err->message);
            g_timeout_add_seconds(10, try_to_connect, NULL);
            goto fail;
        }
        dbus_connection_set_exit_on_disconnect(dbus_g_connection_get_connection(conn), FALSE);
        connproxy = dbus_g_proxy_new_for_name(conn, "org.freedesktop.DBus", "/org/freedesktop/DBus", "org.freedesktop.DBus");
        if(!connproxy){
            dbus_g_connection_unref(conn); conn = NULL;
            warn(connect_err_level, "Failed to create dbus proxy");
            g_timeout_add_seconds(10, try_to_connect, NULL);
            goto fail;
        }
        g_signal_connect(G_OBJECT(connproxy), "destroy", G_CALLBACK(destroy_signaled), NULL);
        dbus_g_proxy_add_signal(connproxy, "NameOwnerChanged", G_TYPE_STRING, G_TYPE_STRING, G_TYPE_STRING, G_TYPE_INVALID);
        dbus_g_proxy_connect_signal(connproxy, "NameOwnerChanged", G_CALLBACK(name_owner_changed), NULL, NULL);
        // XXX: This doesn't actually have any effect, since dbus-glib
        // automatically listens on ALL signals on the proxy.
        ok = dbus_g_proxy_call(connproxy, "AddMatch", &err,
                G_TYPE_STRING, "type='signal',interface=org.freedesktop.DBus,member=NameOwnerChanged,arg0='org.freedesktop.UDisks'",
                G_TYPE_INVALID,
                G_TYPE_INVALID);
        ok = ok && dbus_g_proxy_call(connproxy, "AddMatch", &err,
                G_TYPE_STRING, "type='signal',sender='org.freedesktop.DBus',path='/org/freedesktop/DBus',interface='org.freedesktop.DBus',member='NameOwnerChanged',arg0='org.freedesktop.DeviceKit.Disks'",
                G_TYPE_INVALID,
                G_TYPE_INVALID);
        if(!ok){
            g_signal_handlers_disconnect_by_func(connproxy, G_CALLBACK(destroy_signaled), NULL);
            g_object_unref(connproxy); connproxy = NULL;
            dbus_g_connection_unref(conn); conn = NULL;
            warn(connect_err_level, "Failed to add DBus signal listeners");
            g_timeout_add_seconds(10, try_to_connect, NULL);
            goto fail;
        }
    }

    // Try udisks first
    warn(DEBUG_INFO, "Trying to connect to UDisks");
    propproxy = dbus_g_proxy_new_for_name(conn, "org.freedesktop.UDisks", "/org/freedesktop/UDisks", "org.freedesktop.DBus.Properties");
    if(!propproxy){
        ok = FALSE;
    } else {
        ok = dbus_g_proxy_call(propproxy, "Get", &err,
                G_TYPE_STRING, "org.freedesktop.UDisks",
                G_TYPE_STRING, "DaemonVersion",
                G_TYPE_INVALID,
                G_TYPE_VALUE, &value,
                G_TYPE_INVALID);
    }
    if(ok){
        if(!G_VALUE_HOLDS(&value, G_TYPE_STRING)){
            warn(connect_err_level, "Invalid return type from UDisks DaemonVersion query");
            g_timeout_add_seconds(10, try_to_connect, NULL);
            goto fail;
        }
        ver = g_value_get_string(&value);
        warn(DEBUG_INFO, "Connected to UDisks version %s", ver);

        daemon = g_malloc(8 + strlen(ver));
        sprintf(daemon, "UDisks-%s", ver);
        dbus_name = "org.freedesktop.UDisks";
        path = "/org/freedesktop/UDisks";
        dbus_dev_interface = "org.freedesktop.UDisks.Device";

        if(g_str_has_prefix(ver, "1.0.")) daemon_version = UDISKS_1_0;
        if(!daemon_version) die("UDisks version %s is not supported", ver);

        g_value_unset(&value);
        goto win;
    } else {
        warn(DEBUG_INFO, "Connection to UDisks failed: %s", err?err->message:"<unknown error>");
        if(err){ g_error_free(err); err = NULL; }
    }
    if(propproxy) g_object_unref(propproxy); propproxy=NULL;

    // No udisks, try devicekit-disks
    warn(DEBUG_INFO, "Trying to connect to DeviceKit-Disks");
    propproxy = dbus_g_proxy_new_for_name(conn, "org.freedesktop.DeviceKit.Disks", "/org/freedesktop/DeviceKit/Disks", "org.freedesktop.DBus.Properties");
    if(!propproxy){
        ok = FALSE;
    } else {
        ok = dbus_g_proxy_call(propproxy, "Get", &err,
                G_TYPE_STRING, "org.freedesktop.DeviceKit.Disks",
                G_TYPE_STRING, "DaemonVersion",
                G_TYPE_INVALID,
                G_TYPE_VALUE, &value,
                G_TYPE_INVALID);
    }
    if(ok){
        if(!G_VALUE_HOLDS(&value, G_TYPE_STRING)){
            warn(connect_err_level, "Invalid return type from UDisks DaemonVersion query");
            g_timeout_add_seconds(10, try_to_connect, NULL);
            goto fail;
        }
        ver = g_value_get_string(&value);
        warn(DEBUG_INFO, "Connected to DeviceKit-Disks version %s", ver);

        daemon = g_malloc(17 + strlen(ver));
        sprintf(daemon, "DeviceKit-Disks-%s", ver);
        dbus_name = "org.freedesktop.DeviceKit.Disks";
        path = "/org/freedesktop/DeviceKit/Disks";
        dbus_dev_interface = "org.freedesktop.DeviceKit.Disks.Device";

        if(!strcmp(ver, "009")) daemon_version = DKD_009;
        if(!daemon_version) die("DeviceKit-Disks version %s is not supported", ver);

        g_value_unset(&value);
        goto win;
    } else {
        warn(DEBUG_INFO, "Connection to DeviceKit-Disks failed: %s", err?err->message:"<unknown error>");
        if(err){ g_error_free(err); err = NULL; }
    }
    if(propproxy) g_object_unref(propproxy); propproxy=NULL;

    // Fail
fail: 
    if(err){ g_error_free(err); err = NULL; }
    while(num_all_volumes>0) rm_volume(all_volumes[0]);
    if(proxy){
        g_signal_handlers_disconnect_by_func(proxy, G_CALLBACK(destroy_signaled), NULL);
        g_object_unref(proxy); proxy=NULL;
    }
    if(propproxy) g_object_unref(propproxy); propproxy=NULL;
    if(daemon) g_free(daemon);
    if(udisks_daemon){
        g_free((void *)udisks_daemon);
        udisks_daemon = NULL;
        udisks_connection_changed(NULL);
    }
    daemon_version = 0;
    connect_err_level = DEBUG_DEBUG;
    in_connect = FALSE;
    return FALSE;

win:
    warn(DEBUG_DEBUG, "Creating proxy for %s", dbus_name);
    proxy = dbus_g_proxy_new_for_name_owner(conn, dbus_name, path, dbus_name, &err);
    if(!proxy){
        warn(connect_err_level, "Failed to create %s proxy: %s", dbus_name, err?err->message:"<unknown error>");
        g_timeout_add_seconds(10, try_to_connect, NULL);
        goto fail;
    }
    g_signal_connect(G_OBJECT(proxy), "destroy", G_CALLBACK(destroy_signaled), NULL);

    if(!load_fs_info()){
        warn(connect_err_level, "Failed to load known filesystems");
        g_timeout_add_seconds(10, try_to_connect, NULL);
        goto fail;
    }

    warn(DEBUG_DEBUG, "Connecting dbus signal handlers");
    dbus_g_proxy_add_signal(proxy, "DeviceAdded", DBUS_TYPE_G_OBJECT_PATH, G_TYPE_INVALID);
    dbus_g_proxy_connect_signal(proxy, "DeviceAdded", G_CALLBACK(device_changed), NULL, NULL);
    dbus_g_proxy_add_signal(proxy, "DeviceChanged", DBUS_TYPE_G_OBJECT_PATH, G_TYPE_INVALID);
    dbus_g_proxy_connect_signal(proxy, "DeviceChanged", G_CALLBACK(device_changed), NULL, NULL);
    dbus_g_proxy_add_signal(proxy, "DeviceRemoved", DBUS_TYPE_G_OBJECT_PATH, G_TYPE_INVALID);
    dbus_g_proxy_connect_signal(proxy, "DeviceRemoved", G_CALLBACK(device_removed), NULL, NULL);

    warn(DEBUG_INFO, "Loading existing devices");
    GPtrArray *devices;
    if(!dbus_g_proxy_call(proxy, "EnumerateDevices", &err,
            G_TYPE_INVALID,
            dbus_g_type_get_collection("GPtrArray", DBUS_TYPE_G_OBJECT_PATH), &devices,
            G_TYPE_INVALID)){
        warn(connect_err_level, "Failed to enumerate devices: %s", err?err->message:"<unknown error>");
        g_timeout_add_seconds(10, try_to_connect, NULL);
        goto fail;
    }

    connect_err_level = DEBUG_ERROR;
    if(udisks_daemon) g_free((void *)udisks_daemon);
    udisks_daemon = daemon; daemon = NULL;
    udisks_connection_changed(udisks_daemon);
    for(int i=0; i<devices->len; i++){
        device_changed(proxy, devices->pdata[i], NULL);
    }
    g_ptr_array_free(devices, TRUE);

    in_connect = FALSE;
    return FALSE;
}

gboolean udisks_init(){
    is_fstab_changed(); // The first call always returns true
    g_timeout_add_seconds(1, poll_blocks, NULL);
    Volumes = g_malloc(max_volumes * sizeof(*Volumes));
    all_volumes = g_malloc(max_all_volumes * sizeof(*all_volumes));

    try_to_connect(NULL);
    return TRUE;
}

struct callback_data {
    const char *action;
    udisks_volume *v;
    udisks_volume *req_v;
    udisks_callback callback;
    int flag;
    gpointer data;
};
static void free_callback_data(gpointer d){
    g_slice_free(struct callback_data, d);
}

static void udisks_mount_cb(DBusGProxy *proxy, DBusGProxyCall *call, gpointer p){
    struct callback_data *d = (struct callback_data *)p;
    GError *err = NULL;
    char *mount_path = NULL;
    gboolean ok = dbus_g_proxy_end_call(proxy, call, &err,
                                        G_TYPE_STRING, &mount_path,
                                        G_TYPE_INVALID);
    if(err){
        warn(DEBUG_INFO, "%s %s: %s", d->action, ok?"succeeded":"failed", err->message);
    } else {
        warn(DEBUG_INFO, "%s %s", d->action, ok?"succeeded":"failed");
    }
    if(ok){
        if(err) g_error_free(err);
        err = NULL;
        g_set_error(&err, 0, 0, "%s mounted on %s", udisks_get_display_name(d->req_v), mount_path);
    }
    d->req_v->pending_actions--;
    if(d->callback) d->callback(d->req_v, mount_path, d->data, !ok, err);
    udisks_unref(d->req_v);
    if(err) g_error_free(err);
}
gboolean udisks_mount(udisks_volume *v, udisks_callback callback, gpointer data){
    const char *fstype = "auto";
    switch(v->type){
      case UDISKS_FILESYSTEM:
        if(v->d.fs.info) fstype=v->d.fs.info->id;
        break;
    }
    warn(DEBUG_INFO, "Mounting %s", v->path);
    struct callback_data *d = g_slice_new(struct callback_data);
    d->action = "Mount";
    d->v = v;
    d->req_v = v;
    d->callback = callback;
    d->data = data;
    udisks_ref(v);
    v->pending_actions++;
    dbus_g_proxy_set_interface(v->proxy, dbus_dev_interface);
    if(dbus_g_proxy_begin_call(v->proxy, "FilesystemMount",
                               udisks_mount_cb, d, free_callback_data,
                               G_TYPE_STRING, fstype,
                               G_TYPE_STRV, NULL,
                               G_TYPE_INVALID)){
        return TRUE;
    } else {
        v->pending_actions--;
        udisks_unref(v);
        g_slice_free(struct callback_data, d);
        return FALSE;
    }
}

static void udisks_noret_cb(DBusGProxy *proxy, DBusGProxyCall *call, gpointer p){
    struct callback_data *d = (struct callback_data *)p;
    GError *err = NULL;
    gboolean ok = dbus_g_proxy_end_call(proxy, call, &err,
                                        G_TYPE_INVALID);
    if(err) warn(DEBUG_INFO, "%s %s: %s", d->action, ok?"succeeded":"failed", err->message);
    else warn(DEBUG_INFO, "%s %s", d->action, ok?"succeeded":"failed");
    d->req_v->pending_actions--;
    if(d->callback) d->callback(d->req_v, NULL, d->data, !ok, err);
    g_free((void *)d->req_v->old_label); d->req_v->old_label = NULL;
    udisks_unref(d->req_v);
    if(err) g_error_free(err);
}

gboolean udisks_unmount(udisks_volume *v, udisks_callback callback, gpointer data){
    warn(DEBUG_INFO, "Unmounting %s", v->path);
    struct callback_data *d = g_slice_new(struct callback_data);
    d->action = "Unmount";
    d->v = v;
    d->req_v = v;
    d->callback = callback;
    d->data = data;
    udisks_ref(v);
    g_free((void *)v->old_label); v->old_label=g_strdup(v->label);
    v->pending_actions++;
    dbus_g_proxy_set_interface(v->proxy, dbus_dev_interface);
    if(dbus_g_proxy_begin_call(v->proxy, "FilesystemUnmount",
                               udisks_noret_cb, d, free_callback_data,
                               G_TYPE_STRV, NULL,
                               G_TYPE_INVALID)){
        return TRUE;
    } else {
        v->pending_actions--;
        g_free((void *)v->old_label); v->old_label = NULL;
        udisks_unref(v);
        g_slice_free(struct callback_data, d);
        return FALSE;
    }
}

gboolean udisks_detach(udisks_volume *v, udisks_callback callback, gpointer data){
    udisks_volume *req_v = v;
    while(!v->can_detach){
        if(!v->parent) return FALSE;
        v=v->parent;
    }
    warn(DEBUG_INFO, "Detaching %s", v->path);
    struct callback_data *d = g_slice_new(struct callback_data);
    d->action = "Detach";
    d->v = v;
    d->req_v = req_v;
    d->callback = callback;
    d->data = data;
    udisks_ref(req_v);
    g_free((void *)req_v->old_label); req_v->old_label=g_strdup(req_v->label);
    req_v->pending_actions++;
    dbus_g_proxy_set_interface(v->proxy, dbus_dev_interface);
    if(dbus_g_proxy_begin_call(v->proxy, "DriveDetach",
                               udisks_noret_cb, d, free_callback_data,
                               G_TYPE_STRV, NULL,
                               G_TYPE_INVALID)){
        return TRUE;
    } else {
        req_v->pending_actions--;
        g_free((void *)req_v->old_label); req_v->old_label = NULL;
        udisks_unref(req_v);
        g_slice_free(struct callback_data, d);
        return FALSE;
    }
}

gboolean udisks_eject(udisks_volume *v, udisks_callback callback, gpointer data){
    udisks_volume *req_v = v;
    while(!v->can_eject){
        if(!v->parent) return FALSE;
        v=v->parent;
    }
    warn(DEBUG_INFO, "Ejecting %s", v->path);
    struct callback_data *d = g_slice_new(struct callback_data);
    d->action = "Eject";
    d->v = v;
    d->req_v = req_v;
    d->callback = callback;
    d->data = data;
    udisks_ref(req_v);
    g_free((void *)req_v->old_label); req_v->old_label=g_strdup(req_v->label);
    req_v->pending_actions++;
    dbus_g_proxy_set_interface(v->proxy, dbus_dev_interface);
    if(dbus_g_proxy_begin_call(v->proxy, "DriveEject",
                               udisks_noret_cb, d, free_callback_data,
                               G_TYPE_STRV, NULL,
                               G_TYPE_INVALID)){
        return TRUE;
    } else {
        req_v->pending_actions--;
        g_free((void *)req_v->old_label); req_v->old_label = NULL;
        udisks_unref(req_v);
        g_slice_free(struct callback_data, d);
        return FALSE;
    }
}

static void udisks_lock_cb(DBusGProxy *proxy, DBusGProxyCall *call, gpointer p){
    struct callback_data *d = (struct callback_data *)p;
    GError *err = NULL;
    gboolean ok = dbus_g_proxy_end_call(proxy, call, &err,
                                        G_TYPE_INVALID);
    if(err) warn(DEBUG_INFO, "%s %s: %s", d->action, ok?"succeeded":"failed", err->message);
    else warn(DEBUG_INFO, "%s %s", d->action, ok?"succeeded":"failed");
    device_changed(NULL, d->v->path, NULL);
    d->req_v->pending_actions--;
    if(d->callback) d->callback(d->req_v, d->v, d->data, !ok, err);
    g_free((void *)d->req_v->old_label); d->req_v->old_label = NULL;
    udisks_unref(d->req_v);
    if(err) g_error_free(err);
}
gboolean udisks_lock(udisks_volume *v, udisks_callback callback, gpointer data){
    udisks_volume *req_v = v;
    while(!v->can_lock){
        if(!v->parent) return FALSE;
        v=v->parent;
    }
    warn(DEBUG_INFO, "Locking %s", v->path);
    struct callback_data *d = g_slice_new(struct callback_data);
    d->action = "Lock";
    d->v = v;
    d->req_v = req_v;
    d->callback = callback;
    d->data = data;
    udisks_ref(req_v);
    g_free((void *)req_v->old_label); req_v->old_label=g_strdup(req_v->label);
    req_v->pending_actions++;
    dbus_g_proxy_set_interface(v->proxy, dbus_dev_interface);
    if(dbus_g_proxy_begin_call(v->proxy, "LuksLock",
                               udisks_lock_cb, d, free_callback_data,
                               G_TYPE_STRV, NULL,
                               G_TYPE_INVALID)){
        return TRUE;
    } else {
        req_v->pending_actions--;
        g_free((void *)req_v->old_label); req_v->old_label = NULL;
        udisks_unref(req_v);
        g_slice_free(struct callback_data, d);
        return FALSE;
    }
}

static void udisks_unlock_cb(DBusGProxy *proxy, DBusGProxyCall *call, gpointer p);
static gboolean udisks_unlock_internal(udisks_volume *v, udisks_callback callback, gpointer data, gboolean did_saved, GError **err){
    struct callback_data *d = g_slice_new(struct callback_data);
    d->action = "Unlock";
    d->v = v;
    d->req_v = v;
    d->callback = callback;
    d->data = data;

    char *pass = NULL;
    if(!did_saved) pass = load_pass_from_keyring(v);
    if(pass){
        warn(DEBUG_INFO, "Unlocking %s with saved passphrase", v->path);
        d->flag = 0;
    } else {
        warn(DEBUG_DEBUG, "Requesting passphrase for %s", v->path);
        pass = ask_for_pass(v, NULL, err);
        if(!pass){
            g_slice_free(struct callback_data, d);
            return FALSE;
        }
        warn(DEBUG_INFO, "Unlocking %s with requested passphrase", v->path);
        d->flag = 1;
    }

    udisks_ref(v);
    v->pending_actions++;
    dbus_g_proxy_set_interface(v->proxy, dbus_dev_interface);
    gboolean ok = !!dbus_g_proxy_begin_call(v->proxy, "LuksUnlock",
                               udisks_unlock_cb, d, free_callback_data,
                               G_TYPE_STRING, pass,
                               G_TYPE_STRV, NULL,
                               G_TYPE_INVALID);
    free_pass(pass);
    if(!ok){
        v->pending_actions--;
        udisks_unref(v);
        g_slice_free(struct callback_data, d);
    }
    return ok;
}
static void udisks_unlock_cb(DBusGProxy *proxy, DBusGProxyCall *call, gpointer p){
    struct callback_data *d = (struct callback_data *)p;
    GError *err = NULL;
    char *new_dev = NULL;
    gboolean ok = dbus_g_proxy_end_call(proxy, call, &err,
                                        DBUS_TYPE_G_OBJECT_PATH, &new_dev,
                                        G_TYPE_INVALID);
    if(err){
        warn(DEBUG_INFO, "%s %s: %s", d->action, ok?"succeeded":"failed", err->message);
    } else {
        warn(DEBUG_INFO, "%s %s", d->action, ok?"succeeded":"failed");
    }
    if(ok){
        if(err) g_error_free(err);
        err = NULL;
        g_set_error(&err, 0, 0, "%s unlocked as %s", udisks_get_display_name(d->v), new_dev);
    } else if(!d->flag){
        if(err) g_error_free(err);
        err = NULL;
        if(udisks_unlock_internal(d->req_v, d->callback, d->data, TRUE, &err)) return;
    }
    udisks_volume *new_v = NULL;
    if(new_dev){
        device_changed(NULL, new_dev, NULL);
        for(int i=0; i<num_all_volumes; i++){
            if(!g_strcmp0(all_volumes[i]->path, new_dev)) new_v = all_volumes[i];
        }
    }
    d->req_v->pending_actions--;
    if(d->callback) d->callback(d->req_v, new_v, d->data, !ok, err);
    udisks_unref(d->req_v);
    if(err) g_error_free(err);
}
gboolean udisks_unlock(udisks_volume *v, udisks_callback callback, gpointer data){
    GError *err = NULL;
    gboolean ok = udisks_unlock_internal(v, callback, data, FALSE, &err);
    if(err) g_error_free(err);
    return ok;
}

gboolean udisks_forget_passphrase(udisks_volume *v, udisks_callback callback, gpointer data){
    udisks_volume *req_v=v;
    for(; v; v=v->parent){
        if(v->can_unlock || v->can_lock){
            warn(DEBUG_INFO, "Forgetting passphrase for %s", v->path);
            forget_pass_from_keyring(v);
            if(callback) callback(req_v, v, data, !TRUE, NULL);
            return TRUE;
        }
    }
    warn(DEBUG_WARN, "Cannot find a crypto device for %s to forget passphrase", v->path);
    return FALSE;
}

gboolean udisks_relabel(udisks_volume *v, udisks_callback callback, gpointer data){
    if(!v->can_relabel || !v->d.fs.info || v->d.fs.info->max_label_len<=0){
        warn(DEBUG_WARN, "Cannot relabel %s", v->path);
        return FALSE;
    }

    char *label = ask_for_label(v, NULL);
    if(!label) return FALSE;

    warn(DEBUG_INFO, "Relabeling %s as %s", v->path, label);
    struct callback_data *d = g_slice_new(struct callback_data);
    d->action = "Relabel";
    d->v = v;
    d->req_v = v;
    d->callback = callback;
    d->data = data;
    udisks_ref(v);
    v->pending_actions++;
    dbus_g_proxy_set_interface(v->proxy, dbus_dev_interface);
    gboolean ok = !!dbus_g_proxy_begin_call(v->proxy, "FilesystemSetLabel",
                               udisks_noret_cb, d, free_callback_data,
                               G_TYPE_STRING, label,
                               G_TYPE_INVALID);
    if(!ok){
        v->pending_actions--;
        udisks_unref(v);
        g_slice_free(struct callback_data, d);
    }
    g_free(label);
    return ok;
}
